@@ -27,6 +27,8 @@ from hy.models.symbol import HySymbol
 from hy.models.list import HyList
 from hy.models.dict import HyDict
 
+from hy.util import flatten_literal_list
+
 import ast
 import sys
 
@@ -63,6 +65,7 @@ class HyASTCompiler(object):
 
     def _mangle_branch(self, tree):
         ret = []
+        tree = list(flatten_literal_list(tree))
         tree.reverse()
 
         if self.returnable and len(tree) > 0:
@@ -72,10 +75,21 @@ class HyASTCompiler(object):
                 ret.append(ast.Return(value=el,
                                       lineno=el.lineno,
                                       col_offset=el.col_offset))
-        ret += [ast.Expr(value=el,
-                         lineno=el.lineno,
-                         col_offset=el.col_offset)
-                if not isinstance(el, ast.stmt) else el for el in tree]  # NOQA
+            if isinstance(el, ast.FunctionDef):
+                ret.append(ast.Return(
+                    value=ast.Name(
+                        arg=el.name, id=el.name, ctx=ast.Load(),
+                        lineno=el.lineno, col_offset=el.col_offset),
+                    lineno=el.lineno, col_offset=el.col_offset))
+
+        for el in tree:
+            if isinstance(el, ast.stmt):
+                ret.append(el)
+                continue
+
+            ret.append(ast.Expr(value=el,
+                       lineno=el.lineno,
+                       col_offset=el.col_offset))
 
         ret.reverse()
         return ret
@@ -85,6 +99,7 @@ class HyASTCompiler(object):
         return [self.compile(x) for x in entries]
 
     @builds("do")
+    @builds("progn")
     def compile_do_expression(self, expr):
         return [self.compile(x) for x in expr[1:]]
 
@@ -133,8 +148,7 @@ class HyASTCompiler(object):
             name = str(name)
         else:
             # Python2 requires an ast.Name, set to ctx Store.
-            name = self.compile(name)
-            name.ctx = ast.Store()
+            name = self._storeize(self.compile(name))
 
         return ast.ExceptHandler(
             lineno=expr.start_line,
@@ -151,8 +165,16 @@ class HyASTCompiler(object):
     @builds("if")
     def compile_if_expression(self, expr):
         expr.pop(0)
-        test = self.compile(expr.pop(0))
-        body = self._code_branch(self.compile(expr.pop(0)))
+        try:
+            test = expr.pop(0)
+        except IndexError:
+            raise TypeError("if expects at least 2 arguments, got 0")
+        test = self.compile(test)
+        try:
+            body = expr.pop(0)
+        except IndexError:
+            raise TypeError("if expects at least 2 arguments, got 1")
+        body = self._code_branch(self.compile(body))
         orel = []
         if len(expr) > 0:
             orel = self._code_branch(self.compile(expr.pop(0)))
@@ -165,7 +187,17 @@ class HyASTCompiler(object):
 
     @builds("print")
     def compile_print_expression(self, expr):
-        expr.pop(0)  # print
+        call = expr.pop(0)  # print
+        if sys.version_info[0] >= 3:
+            call = self.compile(call)
+            # AST changed with Python 3, we now just call it.
+            return ast.Call(
+                keywords=[],
+                func=call,
+                args=[self.compile(x) for x in expr],
+                lineno=expr.start_line,
+                col_offset=expr.start_column)
+
         return ast.Print(
             lineno=expr.start_line,
             col_offset=expr.start_column,
@@ -313,12 +345,7 @@ class HyASTCompiler(object):
     def compile_with_as_expression(self, expr):
         expr.pop(0)  # with-as
         ctx = self.compile(expr.pop(0))
-        thing = self.compile(expr.pop(0))
-        if isinstance(thing, ast.Tuple):
-            for x in thing.elts:
-                x.ctx = ast.Store()
-
-        thing.ctx = ast.Store()
+        thing = self._storeize(self.compile(expr.pop(0)))
 
         ret = ast.With(context_expr=ctx,
                        lineno=expr.start_line,
@@ -340,6 +367,40 @@ class HyASTCompiler(object):
                          col_offset=expr.start_column,
                          ctx=ast.Load())
 
+    @builds("list_comp")
+    def compile_list_comprehension(self, expr):
+        # (list-comp expr (target iter) cond?)
+        expr.pop(0)
+        expression = expr.pop(0)
+        tar_it = iter(expr.pop(0))
+        targets = zip(tar_it, tar_it)
+
+        cond = self.compile(expr.pop(0)) if expr != [] else None
+
+        ret = ast.ListComp(
+            lineno=expr.start_line,
+            col_offset=expr.start_column,
+            elt=self.compile(expression),
+            generators=[])
+
+        for target, iterable in targets:
+            ret.generators.append(ast.comprehension(
+                target=self._storeize(self.compile(target)),
+                iter=self.compile(iterable),
+                ifs=[]))
+
+        if cond:
+            ret.generators[-1].ifs.append(cond)
+
+        return ret
+
+    def _storeize(self, name):
+        if isinstance(name, ast.Tuple):
+            for x in name.elts:
+                x.ctx = ast.Store()
+        name.ctx = ast.Store()
+        return name
+
     @builds("kwapply")
     def compile_kwapply_expression(self, expr):
         expr.pop(0)  # kwapply
@@ -453,7 +514,8 @@ class HyASTCompiler(object):
                         col_offset=expression.start_column)
 
     @builds("def")
-    @builds("set")
+    @builds("setf")
+    @builds("setv")
     def compile_def_expression(self, expression):
         expression.pop(0)  # "def"
         name = expression.pop(0)
@@ -467,12 +529,7 @@ class HyASTCompiler(object):
             what.name = str(name)
             return what
 
-        name = self.compile(name)
-        if isinstance(name, ast.Tuple):
-            for x in name.elts:
-                x.ctx = ast.Store()
-
-        name.ctx = ast.Store()
+        name = self._storeize(self.compile(name))
 
         return ast.Assign(
             lineno=expression.start_line,
@@ -486,13 +543,7 @@ class HyASTCompiler(object):
 
         expression.pop(0)  # for
         name, iterable = expression.pop(0)
-        target = self.compile_symbol(name)
-
-        if isinstance(target, ast.Tuple):
-            for x in target.elts:
-                x.ctx = ast.Store()
-
-        target.ctx = ast.Store()
+        target = self._storeize(self.compile_symbol(name))
         # support stuff like:
         # (for [x [1 2 3 4]
         #       y [a b c d]] ...)
@@ -508,6 +559,26 @@ class HyASTCompiler(object):
         self.returnable = ret_status
         return ret
 
+    @builds("while")
+    def compile_while_expression(self, expr):
+        expr.pop(0)  # "while"
+
+        try:
+            test = expr.pop(0)
+        except IndexError:
+            raise TypeError("while expects at least 2 arguments, got 0")
+        test = self.compile(test)
+
+        if not expr:
+            raise TypeError("while expects a body")
+
+        return ast.While(test=test,
+                         body=self._mangle_branch([
+                             self.compile(x) for x in expr]),
+                         orelse=[],
+                         lineno=expr.start_line,
+                         col_offset=expr.start_column)
+
     @builds(HyList)
     def compile_list(self, expr):
         return ast.List(
@@ -527,23 +598,26 @@ class HyASTCompiler(object):
         name = "_hy_anon_fn_%d" % (self.anon_fn_count)
         sig = expression.pop(0)
 
-        ret = ast.FunctionDef(name=name,
-                              lineno=expression.start_line,
-                              col_offset=expression.start_column,
-                              args=ast.arguments(args=[
-                                  ast.Name(arg=str(x), id=str(x),
-                                           ctx=ast.Param(),
-                                           lineno=x.start_line,
-                                           col_offset=x.start_column)
-                                  for x in sig],
-                                  vararg=None,
-                                  kwarg=None,
-                                  kwonlyargs=[],
-                                  kw_defaults=[],
-                                  defaults=[]),
-                              body=self._code_branch([
-                                  self.compile(x) for x in expression]),
-                              decorator_list=[])
+        ret = ast.FunctionDef(
+            name=name,
+            lineno=expression.start_line,
+            col_offset=expression.start_column,
+            args=ast.arguments(
+                args=[
+                    ast.Name(
+                        arg=str(x), id=str(x),
+                        ctx=ast.Param(),
+                        lineno=x.start_line,
+                        col_offset=x.start_column)
+                    for x in sig],
+                vararg=None,
+                kwarg=None,
+                kwonlyargs=[],
+                kw_defaults=[],
+                defaults=[]),
+            body=self._code_branch([
+                self.compile(x) for x in expression]),
+            decorator_list=[])
 
         self.returnable = ret_status
         return ret
