@@ -20,15 +20,21 @@
 
 from hy.macros import process as mprocess
 
+import hy.mangle
+
 
 MACROS = [
-    "hy.core.bootstrap",  # defn, cond
+    "hy.core.bootstrap",
+    "hy.core.mangles",
 ]
 
 
 def process(tree):
     load_macros()
-    return mprocess(tree)
+    tree = mprocess(tree)
+    for m in hy.mangle.MANGLES:
+        m().mangle(tree)
+    return tree
 
 
 def load_macros():
